@@ -30,6 +30,7 @@
 
 #include	"mdadm.h"
 #include	"xmalloc.h"
+#include	"udev.h"
 
 #include	<sys/wait.h>
 #include	<dirent.h>
@@ -282,11 +283,11 @@ int Incremental(struct mddev_dev *devlist, struct context *c,
 		 * clustering resource agents
 		 */
 		if (info.array.state & (1 << MD_SB_CLUSTERED))
-			goto out;
+			goto out_unlock;
 
 		/* Couldn't find an existing array, maybe make a new one */
 		mdfd = create_mddev(match ? match->devname : NULL, name_to_use, trustworthy,
-				    chosen_name, 0);
+				    chosen_name, 1);
 
 		if (mdfd < 0)
 			goto out_unlock;
@@ -447,7 +448,6 @@ int Incremental(struct mddev_dev *devlist, struct context *c,
 		info.array.working_disks = 0;
 		for (d = sra->devs; d; d=d->next)
 			info.array.working_disks ++;
-
 	}
 	if (strncmp(chosen_name, DEV_MD_DIR, DEV_MD_DIR_LEN) == 0)
 		md_devname = chosen_name + DEV_MD_DIR_LEN;
@@ -464,7 +464,6 @@ int Incremental(struct mddev_dev *devlist, struct context *c,
 	if (is_container(info.array.level)) {
 		char devnm[32];
 		/* Try to assemble within the container */
-		sysfs_uevent(sra, "change");
 		if (!c->export && c->verbose >= 0)
 			pr_err("container %s now has %d device%s\n",
 			       chosen_name, info.array.working_disks,
@@ -476,6 +475,8 @@ int Incremental(struct mddev_dev *devlist, struct context *c,
 		if (st->ss->load_container)
 			rv = st->ss->load_container(st, mdfd, NULL);
 		close(mdfd);
+		udev_unblock();
+		sysfs_uevent(sra, "change");
 		sysfs_free(sra);
 		if (!rv)
 			rv = Incremental_container(st, chosen_name, c, NULL);
@@ -484,6 +485,7 @@ int Incremental(struct mddev_dev *devlist, struct context *c,
 		 * so that it can eg. try to rebuild degraded array */
 		if (st->ss->external)
 			ping_monitor(devnm);
+		udev_unblock();
 		return rv;
 	}
 
@@ -606,7 +608,11 @@ out:
 		close(mdfd);
 	if (policy)
 		dev_policy_free(policy);
-	sysfs_free(sra);
+	udev_unblock();
+	if (sra) {
+		sysfs_uevent(sra, "change");
+		sysfs_free(sra);
+	}
 	return rv;
 out_unlock:
 	map_unlock(&map);
@@ -1561,7 +1567,7 @@ static int Incremental_container(struct supertype *st, char *devname,
 				trustworthy = LOCAL;
 
 			mdfd = create_mddev(match ? match->devname : NULL, ra->name, trustworthy,
-					    chosen_name, 0);
+					    chosen_name, 1);
 
 			if (!is_fd_valid(mdfd)) {
 				pr_err("create_mddev failed with chosen name %s: %s.\n",
@@ -1581,6 +1587,8 @@ static int Incremental_container(struct supertype *st, char *devname,
 		map_free(map);
 		map = NULL;
 		close_fd(&mdfd);
+		udev_unblock();
+		sysfs_uevent(&info, "change");
 	}
 	if (c->export && result) {
 		char sep = '=';
@@ -1607,25 +1615,11 @@ static int Incremental_container(struct supertype *st, char *devname,
 release:
 	map_free(map);
 	sysfs_free(list);
+	udev_unblock();
+	sysfs_uevent(&info, "change");
 	return rv;
 }
 
-static void remove_from_member_array(struct mdstat_ent *memb,
-				    struct mddev_dev *devlist, int verbose)
-{
-	int subfd = open_dev(memb->devnm);
-
-	if (subfd >= 0) {
-		/*
-		 * Ignore the return value because it's necessary
-		 * to handle failure condition here.
-		 */
-		Manage_subdevs(memb->devnm, subfd, devlist, verbose,
-			       0, UOPT_UNDEFINED, 0);
-		close(subfd);
-	}
-}
-
 /**
  * is_devnode_path() - check if the devname passed might be devnode path.
  * @devnode: the path to check.
@@ -1646,25 +1640,82 @@ static bool is_devnode_path(char *devnode)
 	return false;
 }
 
+/**
+ * Incremental_remove_external() - Remove the device from external container.
+ * @device_devnm: block device to remove.
+ * @container_devnm: the parent container
+ * @mdstat: mdstat file content.
+ * @verbose: verbose flag.
+ *
+ * Fail member device in each subarray and remove member device from external container.
+ * The resposibility of removing member disks from external subararys belongs to mdmon.
+ */
+static mdadm_status_t Incremental_remove_external(char *device_devnm, char *container_devnm,
+						  struct mdstat_ent *mdstat, int verbose)
+{
+	mdadm_status_t rv = MDADM_STATUS_SUCCESS;
+	struct mdstat_ent *memb;
+
+	for (memb = mdstat ; memb ; memb = memb->next) {
+		mdadm_status_t ret = MDADM_STATUS_SUCCESS;
+		int state_fd;
+
+		if (!is_container_member(memb, container_devnm))
+			continue;
+
+		/*
+		 * Checking mdstat is pointles because it might be outdated, try open descriptor
+		 * instead. If it fails, we are fine with that, device is already gone.
+		 */
+		state_fd = sysfs_open_memb_attr(memb->devnm, device_devnm, "state", O_RDWR);
+		if (!is_fd_valid(state_fd))
+			continue;
+
+		ret = sysfs_set_memb_state_fd(state_fd, MEMB_STATE_FAULTY, NULL);
+		if (ret && verbose >= 0)
+			pr_err("Cannot fail member device %s in external subarray %s.\n",
+				device_devnm, memb->devnm);
+
+		close_fd(&state_fd);
+
+		/*
+		 * Don't remove member device from container if it failed to remove it
+		 * from any member array.
+		 */
+		rv |= ret;
+	}
+
+	if (rv == MDADM_STATUS_SUCCESS)
+		rv = sysfs_set_memb_state(container_devnm, device_devnm, MEMB_STATE_REMOVE);
+
+	if (rv && verbose >= 0)
+		pr_err("Cannot remove member device %s from container %s.\n", device_devnm,
+		       container_devnm);
+
+	return rv;
+}
+
 /**
  * Incremental_remove() - Remove the device from all raid arrays.
  * @devname: the device we want to remove, it could be kernel device name or devnode.
  * @id_path: optional, /dev/disk/by-path path to save for bare scenarios support.
  * @verbose: verbose flag.
  *
- * First, fail the device (if needed) and then remove the device from native raid array or external
- * container.  If it is external container, the device is removed from each subarray first.
+ * First, fail the device (if needed) and then remove the device. This code is critical for system
+ * funtionality and that is why it is keept as simple as possible. We do not load devices using
+ * sysfs_read() because any unerelated failure may lead us to abort. We also do not call
+ * Manage_Subdevs().
  */
 int Incremental_remove(char *devname, char *id_path, int verbose)
 {
+	mdadm_status_t rv = MDADM_STATUS_SUCCESS;
 	char *devnm = basename(devname);
-	struct mddev_dev devlist = {0};
 	char buf[SYSFS_MAX_BUF_SIZE];
 	struct mdstat_ent *mdstat;
 	struct mdstat_ent *ent;
 	struct mdinfo mdi;
-	int rv = 1;
-	int mdfd;
+	int mdfd = -1;
+	int retry = 25;
 
 	if (strcmp(devnm, devname) != 0)
 		if (!is_devnode_path(devname)) {
@@ -1693,15 +1744,30 @@ int Incremental_remove(char *devname, char *id_path, int verbose)
 
 	mdfd = open_dev_excl(ent->devnm);
 	if (is_fd_valid(mdfd)) {
+		char *array_state_file = "array_state";
+
+		/**
+		 * This is a workaround for the old issue.
+		 * Incremental_remove() triggered from udev rule when disk is removed from OS
+		 * tries to set array in auto-read-only mode. This can interrupt rebuild
+		 * process which is started automatically, e.g. if array is mounted and
+		 * spare disk is available (I/O errors limit might be achieved faster than disk is
+		 * removed by mdadm). Prevent Incremental_remove() from setting array
+		 * into "auto-read-only", by requiring  exclusive open to succeed.
+		 */
 		close_fd(&mdfd);
-		if (sysfs_get_str(&mdi, NULL, "array_state",
-				  buf, sizeof(buf)) > 0) {
-			if (strncmp(buf, "active", 6) == 0 ||
-			    strncmp(buf, "clean", 5) == 0)
-				sysfs_set_str(&mdi, NULL,
-					      "array_state", "read-auto");
+
+		if (sysfs_get_str(&mdi, NULL, array_state_file, buf, sizeof(buf)) > 0) {
+			char *str_read_auto = map_num_s(sysfs_array_states, ARRAY_READ_AUTO);
+			char *str_active = map_num_s(sysfs_array_states, ARRAY_ACTIVE);
+			char *str_clean = map_num_s(sysfs_array_states, ARRAY_CLEAN);
+
+			if (strncmp(buf, str_active, strlen(str_active)) == 0 ||
+			    strncmp(buf, str_clean, strlen(str_clean)) == 0)
+				sysfs_set_str(&mdi, NULL, array_state_file, str_read_auto);
 		}
 	}
+
 	mdfd = open_dev(ent->devnm);
 	if (mdfd < 0) {
 		if (verbose >= 0)
@@ -1711,38 +1777,46 @@ int Incremental_remove(char *devname, char *id_path, int verbose)
 
 	if (id_path) {
 		struct map_ent *map = NULL, *me;
+
 		me = map_by_devnm(&map, ent->devnm);
 		if (me)
 			policy_save_path(id_path, me);
 		map_free(map);
 	}
 
-	devlist.devname = devnm;
-	devlist.disposition = 'I';
-	/* for a container, we must fail each member array */
 	if (is_mdstat_ent_external(ent)) {
-		struct mdstat_ent *memb;
-		for (memb = mdstat ; memb ; memb = memb->next) {
-			if (is_container_member(memb, ent->devnm))
-				remove_from_member_array(memb,
-					&devlist, verbose);
-		}
-	} else {
-		/*
-		 * This 'I' incremental remove is a try-best effort,
-		 * the failure condition can be safely ignored
-		 * because of the following up 'r' remove.
-		 */
-		Manage_subdevs(ent->devnm, mdfd, &devlist,
-			       verbose, 0, UOPT_UNDEFINED, 0);
+		rv = Incremental_remove_external(devnm, ent->devnm, mdstat, verbose);
+		goto out;
 	}
 
-	devlist.disposition = 'r';
-	rv = Manage_subdevs(ent->devnm, mdfd, &devlist,
-			    verbose, 0, UOPT_UNDEFINED, 0);
+	/* Native arrays are handled separatelly to provide more detailed error handling */
+	rv = sysfs_set_memb_state(ent->devnm, devnm, MEMB_STATE_FAULTY);
+	if (rv) {
+		if (verbose >= 0)
+			pr_err("Cannot fail member device %s in array %s.\n", devnm, ent->devnm);
+		goto out;
+	}
+
+	/*
+	 * If resync/recovery is running, sync thread is interrupted by setting member faulty.
+	 * And it needs to wait some time to let kernel to reap sync thread. If not, it will
+	 * fail to remove it.
+	 */
+	while (retry) {
+		rv = sysfs_set_memb_state(ent->devnm, devnm, MEMB_STATE_REMOVE);
+		if (rv) {
+			sleep_for(0, MSEC_TO_NSEC(200), true);
+			retry--;
+			continue;
+		}
+		break;
+	}
+
+	if (rv && verbose >= 0)
+		pr_err("Cannot remove member device %s from %s.\n", devnm, ent->devnm);
 
-	close_fd(&mdfd);
 out:
+	close_fd(&mdfd);
 	free_mdstat(mdstat);
 	return rv;
 }
